from .classification import (GaussianProcessClassifier,
                             SparseGaussianProcessClassifier)
from .regression import (GaussianProcessRegressor,
                         SparseGaussianProcessRegressor)

__all__ = ['GaussianProcessClassifier',
           'GaussianProcessRegressor',
           'SparseGaussianProcessClassifier',
           'SparseGaussianProcessRegressor']



"""Classes for regression based on `GPy` regression"""

from sklearn.base import RegressorMixin
from GPy.models import GPRegression as _GPRegression
from GPy.models import SparseGPRegression as _SparseGPRegression
from GPy.models import GPHeteroscedasticRegression as _GPHeteroscedasticRegression
from .base import _BaseGP


class GaussianProcessRegressor(_BaseGP, RegressorMixin):
    """Gaussian Process Regressor based on `GPy` regressor.

    Parameters
    ----------
    kernels : str or list of str, optional
        Kernel for `GPy` model.
        If string, that kernel should be in `GPy.kern`.
        If list of str, the sum of kernels is used.
        Default="RBF"
    kernel_options : dict or list of dict, optional
        Kernel options to be set for kernels.
        If `kernels` is str, `kernel_options` should be dict.
        Default={'input_dim': X.shape[1]}
    Y_metadata : optional
        Metadata assosiated with points.
    normalizer : object, optional
        Normalize the outputs Y. Prediction will be un-normalized using
        this normalizer. If normalizer is None, we will normalize using
        ``GPy.util.normalizer.Standardize``. If normalizer is False, no
        normalization will be done.
    noise_var : float, optional
        The noise variance for Gaussian likelhood, defaults to 1.
    mean_function : optional
    method : {'optimize', 'optimize_restarts'}, optional
        Invokes passed method to fit `GPy` model. 
        For 'optimize_restarts' perform random restarts of the
        model, and set the model to the best.

    Attributes
    ----------
    estimator_ : object
        `GPy` estimator fitted to data.
    n_features_ : int
        Number of features in fitted data.
    """

    def __init__(self, kernels=None, kernel_options=None, Y_metadata=None,
                 normalizer=True, noise_var=1.0, mean_function=None,
                 method='optimize'):
        self.normalizer = normalizer
        self.noise_var = noise_var
        self.kernels = kernels
        self.kernel_options = kernel_options
        self.Y_metadata = Y_metadata
        self.mean_function = mean_function
        self.method = method

    def _get_model(self, X, y, kernel):
        """Returns the `GPy` regression model with initialized params

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training set.
        y : array-like, shape (n_samples)
            Target values.
        kernel : object
            `GPy` kernel.

        Returns
        -------
        model : object
            Returns the `GPy` regression model.

        """
        return _GPRegression(X, y, kernel, self.Y_metadata, self.normalizer,
                             self.noise_var, self.mean_function)


class SparseGaussianProcessRegressor(_BaseGP, RegressorMixin):
    """Sparse Gaussian Process Regressor based on `GPy` sparse regressor.

    Parameters
    ----------
    kernels : str or list of str, optional
        Kernel for `GPy` model.
        If string, that kernel should be in `GPy.kern`.
        If list of str, the sum of kernels is used.
        Default="RBF"
    kernel_options : dict or list of dict, optional
        Kernel options to be set for kernels.
        If `kernels` is str, `kernel_options` should be dict.
        Default={'input_dim': X.shape[1]}
    Z : array-like, shape (num_inducing, input_dim), optional
        Inducing inputs
    num_inducing : int, optional (default=10)
        Number of inducing points (Ignored if Z is not None)
        If specified select inducing points randomly. 
    Y_metadata : optional
        Metadata assosiated with points.
    X_variance : array-like, optional
        Input uncertainties, one per input X
    normalizer : object, optional
        Normalize the outputs Y. Prediction will be un-normalized using
        this normalizer. If normalizer is None, we will normalize using
        ``GPy.util.normalizer.Standardize``. If normalizer is False, no
        normalization will be done.
    mean_function : optional
    method : {'optimize', 'optimize_restarts'}, optional
        Invokes passed method to fit `GPy` model. 
        For 'optimize_restarts' perform random restarts of the
        model, and set the model to the best.

    Attributes
    ----------
    estimator_ : object
        `GPy` estimator fitted to data.
    n_features_ : int
        Number of features in fitted data.
    """

    def __init__(self, kernels=None, kernel_options=None, Z=None,
                 num_inducing=10, Y_metadata=None, X_variance=None,
                 normalizer=True, mean_function=None, method='optimize'):
        self.kernels = kernels
        self.kernel_options = kernel_options
        self.Z = Z
        self.num_inducing = num_inducing
        self.Y_metadata = Y_metadata
        self.X_variance = X_variance
        self.normalizer = normalizer
        self.mean_function = mean_function
        self.method = method

    def _get_model(self, X, y, kernel):
        """Returns the `GPy` regression model with initialized params

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training set.
        y : array-like, shape (n_samples)
            Target values.
        kernel : object
            `GPy` kernel.

        Returns
        -------
        model : object
            Returns the `GPy` regression model.

        """
        return _SparseGPRegression(X, y, kernel=kernel, Z=self.Z,
                                   num_inducing=self.num_inducing,
                                   X_variance=self.X_variance,
                                   mean_function=self.mean_function,
                                   normalizer=self.normalizer)

# class GaussianProcessHeteroscedasticRegressor(_BaseGP, RegressorMixin):
#     """Heteroscedastic Gaussian Process Regressor based on `GPy` regressor.

#     Parameters
#     ----------
#     kernels : str or list of str, optional
#         Kernel for `GPy` model.
#         If string, that kernel should be in `GPy.kern`.
#         If list of str, the sum of kernels is used.
#         Default="RBF"
#     kernel_options : dict or list of dict, optional
#         Kernel options to be set for kernels.
#         If `kernels` is str, `kernel_options` should be dict.
#         Default={'input_dim': X.shape[1]}
#     Y_metadata : optional
#         Metadata assosiated with points.
#     method : {'optimize', 'optimize_restarts'}, optional
#         Invokes passed method to fit `GPy` model. 
#         For 'optimize_restarts' perform random restarts of the
#         model, and set the model to the best.

#     Attributes
#     ----------
#     estimator_ : object
#         `GPy` estimator fitted to data.
#     n_features_ : int
#         Number of features in fitted data.
#     """
#     def __init__(self, kernels=None, kernel_options=None, Y_metadata=None,
#                  method='optimize'):
#         self.kernels = kernels
#         self.kernel_options = kernel_options
#         self.Y_metadata = Y_metadata
#         self.method = method

#     def _get_model(self, X, y, kernel):
#         """Returns the `GPy` regression model with initialized params

#         Parameters
#         ----------
#         X : array-like, shape (n_samples, n_features)
#             Training set.
#         y : array-like, shape (n_samples)
#             Target values.
#         kernel : object
#             `GPy` kernel.

#         Returns
#         -------
#         model : object
#             Returns the `GPy` regression model.
#         """
#         return _GPHeteroscedasticRegression(X, y, kernel, 
#                                             Y_metadata=self.Y_metadata)



"""Base class for estimators from `GPy` to use them as sklearn estimators"""

import numpy as np
from sklearn.utils import check_X_y, assert_all_finite, check_array
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.multiclass import check_classification_targets
from sklearn.base import BaseEstimator
from copy import deepcopy
from GPy import kern as gpy_kern
from GPy import load as load_gp
from abc import ABCMeta, abstractmethod
import six


class _BaseGP(six.with_metaclass(ABCMeta, BaseEstimator)):
    """Base class for all estimators in EDR-GP

    Parameters
    ----------
    kernels : str or list of str or object, optional
        Kernel for `GPy` model.
        If string, that kernel should be in `GPy.kern`.
        If list of str, the sum of kernels is used.
        It is also possible to use GPy.kern objects.
        Default="RBF".
    kernel_options : dict or list of dict, optional
        Kernel options to be set for kernels.
        If `kernels` is str, `kernel_options` should be dict.
        Default={'input_dim': X.shape[1]}.
    Y_metadata : optional
        Metadata assosiated with points.
    mean_function : optional
    method : {'optimize', 'optimize_restarts'}, optional
        Invokes passed method to fit `GPy` model. 
        For 'optimize_restarts' perform random restarts of the
        model, and set the model to the best.

    Attributes
    ----------
    estimator_ : object
        `GPy` estimator fitted to data.
    n_features_ : int
        Number of features in fitted data.
    """

    def fit(self, X, y, **opt_kws):
        """Fit the model according to the given training data

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training set.
        y : array-like, shape (n_samples)
            Target values.

        Returns
        -------
        self : object
            Returns self.
        """
        X, y = self._check_data(X, y)

        self.n_features_ = X.shape[1]
        kernel = self._make_kernel()
        self.estimator_ = self._get_model(X, y, kernel)

        opt_kws.setdefault('messages', False)
        opt_kws.setdefault('max_iters', 1000)
        getattr(self.estimator_, self.method)(**opt_kws)
        return self

    def _check_data(self, X, y):
        """Check data before fitting the model

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training set.
        y : array-like, shape (n_samples)
            Target values.

        Returns
        -------
        X : ndarray, shape (n_samples, n_features)
        y : ndarray, shape (n_samples, 1)
        """
        X, y = check_X_y(X, y, accept_sparse=False)
        if self._estimator_type == 'classifier':
            check_classification_targets(y)
        y = y[:, np.newaxis]
        return X, y

    def _check_input(self, X):
        """Check X before predicting

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Test points.

        Returns
        -------
        X : ndarray, shape (n_samples, n_features)
        """
        X = check_array(X, accept_sparse=False)
        if X.shape[1] != self.n_features_:
            raise ValueError("X has {} features per sample; expecting {}"
                             .format(X.shape[1], self.n_features_))
        return X

    def _make_kernel(self):
        """Create kernel for `GPy` model.

        Returns
        -------
        kernel : object
            Returns `GPy` kernel.
        """

        # kernel will be initiated as 'RBF' in model automatically
        if self.kernels is None:
            return self.kernels

        if hasattr(self.kernels, '__module__'):
            if self.kernels.__module__.startswith('GPy.kern'):
                return self.kernels

        if isinstance(self.kernels, str):
            self.kernels = [self.kernels]

        if isinstance(self.kernel_options, dict):
            self.kernel_options = [self.kernel_options]

        kernels = [getattr(gpy_kern, kern) for kern in self.kernels]
        input_dim = {'input_dim': self.n_features_}

        if self.kernel_options is None:
            options = [input_dim] * len(self.kernels)
        elif len(kernels) == len(self.kernel_options):
            options = deepcopy(self.kernel_options)
            for opt in options:
                opt.update(input_dim)
        else:
            raise(ValueError)

        kernel = np.sum([kern(**opt) for kern, opt in zip(kernels, options)])
        return kernel

    @abstractmethod
    def _get_model(self, X, y, kenrel):
        pass

    def _check_predict(self, X):
        """Check X and model before predicting

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Testing data.

        Returns
        -------
        X : ndarray, shape (n_samples, n_features)
        """
        X = self._check_input(X)
        check_is_fitted(self, 'estimator_')
        return X

    def predict(self, X):
        """Predict the target for the new points

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Testing points.

        Returns
        -------
        y_proba : ndarray
            For classification returns the posterior probability of the
            sample for second class in the model.
            For regression returns the posterior mean for each point 
            of the sample.

        """
        X = self._check_predict(X)
        y_pred = self.estimator_.predict(X)[0][:, 0]
        assert_all_finite(y_pred)
        return y_pred

    def predict_variance(self, X):
        """Predict the target for the new points.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Testing points.

        Returns
        -------
        y_proba : ndarray
            Returns the posterior variance for each point of the sample.

        """
        X = self._check_predict(X)
        return self.estimator_.predict(X)[1]

    def predict_gradient(self, X):
        """Compute the derivative of the predicted latent function.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Testing points.

        Returns
        -------
        grads : ndarray, shape (n_samples, n_features)
            Returns the gradients of the sample.
        """
        X = self._check_predict(X)
        return self.estimator_.predictive_gradients(X)[0][:, :, 0]

    def save(self, model_path):
        """Save model to `model_path` file with '.pickle' extension

        Parameters
        ----------
        model_path: str
            Path to which model will be dumped. The extension '.pickle'
            will be add to path automatically

        Returns
        -------
        None
        """
        if not model_path.endswith('.pickle'):
            model_path += '.pickle'
        self.estimator_.pickle(model_path)


    def load(self, model_path):
        """Load model from `model_path`

        Parameters
        ----------
        model_path: str
            Path to which model was dumped. The extension '.pickle'
            will be add to path automatically.

        Returns
        -------
        None
        """
        if not model_path.endswith('.pickle'):
            model_path += '.pickle'
        self.estimator_ = load_gp(model_path)



"""Class for classification based on `GPy` classification"""

from sklearn.base import ClassifierMixin
from GPy.models import GPClassification as _GPClassification
from GPy.models import SparseGPClassification as _SGPClassification
from .base import _BaseGP


class GaussianProcessClassifier(_BaseGP, ClassifierMixin):
    """GaussianProcessClassifier based on `GPy` classifier

    Parameters
    ----------
    kernels : str or list of str, optional (default="RBF")
        Kernel for `GPy` model.
        If string, that kernel should be in `GPy.kern`.
        If list of str, the sum of kernels is used.
    kernel_options : dict or list of dict, optional
        Kernel options to be set for kernels.
        If `kernels` is str, `kernel_options` should be dict.
        Default={'input_dim': X.shape[1]}
    Y_metadata : optional
        Metadata assosiated with points.
    mean_function : optional
    method : {'optimize', 'optimize_restarts'}, optional
        Invokes passed method to fit `GPy` model. 
        For 'optimize_restarts' perform random restarts of the
        model, and set the model to the best.

    Attributes
    ----------
    estimator_ : object
        `GPy` estimator fitted to data.
    n_features_ : int
        Number of features in fitted data.
    """

    def __init__(self, kernels=None, kernel_options=None, Y_metadata=None,
                 mean_function=None, method='optimize'):
        self.kernels = kernels
        self.kernel_options = kernel_options
        self.Y_metadata = Y_metadata
        self.mean_function = mean_function
        self.method = method

    def _get_model(self, X, y, kernel):
        """Returns the `GPy` classification model with initialized params

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training set.
        y : array-like, shape (n_samples)
            Target values.
        kernel : object
            `GPy` kernel.

        Returns
        -------
        model : object
            Returns the `GPy` classification model.
        """
        return _GPClassification(X, y, kernel, self.Y_metadata,
                                 self.mean_function)

    def predict(self, X):
        """Predict the target for the new points

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Testing points.

        Returns
        -------
        y : ndarray
            Returns 1 if the posterior probability of second class is higher
            than 0.5, else returns 0.

        """
        return super(GaussianProcessClassifier, self).predict(X) > 0.5

    def predict_proba(self, X):
        """Predict the target for the new points

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Testing points.

        Returns
        -------
        y_proba : ndarray
            Returns the posterior probability of the
            sample for second class in the model.

        """
        return super(GaussianProcessClassifier, self).predict(X)


class SparseGaussianProcessClassifier(_BaseGP, ClassifierMixin):
    """SparseGaussianProcessClassifier based on `GPy` sparse classifier

    Parameters
    ----------
    kernels : str or list of str, optional (default="RBF")
        Kernel for `GPy` model.
        If string, that kernel should be in `GPy.kern`.
        If list of str, the sum of kernels is used.
    kernel_options : dict or list of dict, optional
        Kernel options to be set for kernels.
        If `kernels` is str, `kernel_options` should be dict.
        Default={'input_dim': X.shape[1]}
    likelihood : object
        GPy likelihood, defaults to Binomial with probit link_function
    Z : array-like, shape (num_inducing, input_dim), optional
        Inducing inputs
    num_inducing : int, optional (default=10)
        Number of inducing points (Ignored if Z is not None)
        If specified select inducing points randomly. 
    Y_metadata : optional
        Metadata assosiated with points.
    method : {'optimize', 'optimize_restarts'}, optional
        Invokes passed method to fit `GPy` model. 
        For 'optimize_restarts' perform random restarts of the
        model, and set the model to the best.

    Attributes
    ----------
    estimator_ : object
        `GPy` estimator fitted to data.
    n_features_ : int
        Number of features in fitted data.
    """

    def __init__(self, kernels=None, kernel_options=None, likelihood=None,
                 Z=None, num_inducing=10, Y_metadata=None, method='optimize'):
        self.kernels = kernels
        self.kernel_options = kernel_options
        self.likelihood = likelihood
        self.Z = Z
        self.num_inducing = num_inducing
        self.Y_metadata = Y_metadata
        self.method = method

    def _get_model(self, X, y, kernel):
        return _SGPClassification(X, y, likelihood=self.likelihood, kernel=kernel,
                           Z=self.Z, num_inducing=self.num_inducing,
                           Y_metadata=self.Y_metadata)

    def predict(self, X):
        """Predict the target for the new points

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Testing points.

        Returns
        -------
        y : ndarray
            Returns 1 if the posterior probability of second class is higher
            than 0.5, else returns 0.

        """
        return super(SparseGaussianProcessClassifier, self).predict(X) > 0.5

    def predict_proba(self, X):
        """Predict the target for the new points

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Testing points.

        Returns
        -------
        y_proba : ndarray
            Returns the posterior probability of the
            sample for second class in the model.

        """
        return super(SparseGaussianProcessClassifier, self).predict(X)



import pytest
import numpy as np
from copy import deepcopy
from edrgp.gp_model.regression import GaussianProcessRegressor
from edrgp.gp_model.regression import SparseGaussianProcessRegressor
from edrgp.edr import EffectiveDimensionalityReduction
from edrgp.base import BlockEDR
from edrgp.datasets import (get_gaussian_inputs,
                            get_tanh_targets,
                            get_edr_target,
                            get_beta_inputs)
from sklearn.feature_selection import mutual_info_regression
from sklearn.decomposition import PCA, SparsePCA
from sklearn.preprocessing import StandardScaler
from edrgp.utils import SVDTransformer
from edrgp.utils import discrepancy
from scipy.sparse import random as random_sparse
import GPy


def get_2d_data(mean=None):
    '''Returns a simple 2D dataset'''
    if mean is None:
        mean = [0, 0]
    X = get_gaussian_inputs(
        eig_values=[1, 0.3], sample_size=500,
        eig_vectors=np.array([[1, 1], [-1, 1]]),
        mean=mean)
    y = get_tanh_targets(X, [0.5, 0.5])
    return X, y


def test_sparse_regression():
    np.random.seed(101)
    N = 50
    noise_var = 0.05

    X = np.linspace(0, 10, 50)[:, None]
    k = GPy.kern.RBF(1)
    y = np.random.multivariate_normal(
        np.zeros(N), k.K(X)+np.eye(N)*np.sqrt(noise_var)).reshape(-1, 1)

    gp = GaussianProcessRegressor()
    gp.fit(X, y)

    sgp = SparseGaussianProcessRegressor(num_inducing=12)
    sgp.fit(X, y)

    assert abs(gp.estimator_.log_likelihood() -
               sgp.estimator_.log_likelihood()[0][0]) < 0.5


@pytest.mark.parametrize("mean", [[0, 0], [10, -10]])
def test_mi(mean):
    X, y = get_2d_data(mean)
    edr = EffectiveDimensionalityReduction(GaussianProcessRegressor(),
                                           SVDTransformer(), n_components=1,
                                           normalize=True)
    edr.fit(X, y)
    mi = mutual_info_regression(edr.transform(X), y)[0]
    assert mi > 1


@pytest.mark.parametrize("normalize", [True, False])
def test_translation(normalize):
    X, y = get_2d_data(mean=[10, -10])
    edr = EffectiveDimensionalityReduction(GaussianProcessRegressor(),
                                           SVDTransformer(), n_components=1,
                                           normalize=normalize)
    edr.fit(X, y)
    components_shift = edr.components_

    X -= X.mean(0)
    edr2 = deepcopy(edr)
    edr2.fit(X, y)
    components_no_shift = edr2.components_
    assert np.allclose(components_shift, components_no_shift, rtol=1e-3)


@pytest.mark.parametrize("mean", [[0, 0, 0, 0], [10, -10, 100, -100]])
def test_preprocess(mean):
    X = get_gaussian_inputs(
        eig_values=[1, 0.3, 0.001, 0.001], sample_size=500, mean=mean)
    y = get_tanh_targets(X, [0.5, 0.5, 0, 0])

    edr = EffectiveDimensionalityReduction(GaussianProcessRegressor(),
                                           SVDTransformer(), n_components=1,
                                           normalize=True,
                                           preprocessor=PCA(n_components=2))
    edr.fit(X, y)
    components_shift = edr.components_

    X -= X.mean(0)
    edr = EffectiveDimensionalityReduction(GaussianProcessRegressor(),
                                           SVDTransformer(), n_components=1,
                                           normalize=True,
                                           preprocessor=PCA(n_components=2))
    edr.fit(X, y)
    components_no_shift = edr.components_
    assert np.allclose(components_shift, components_no_shift, rtol=1e-3)


@pytest.mark.parametrize("mean", [[0, 0], [10, -10]])
def test_scaling(mean):
    X, y = get_2d_data(mean)
    # EDR with scaling
    edr_sc = EffectiveDimensionalityReduction(GaussianProcessRegressor(),
                                              SVDTransformer(), normalize=True)
    edr_sc.fit(X, y)
    x1 = edr_sc.transform(X-np.mean(X, axis=0))
    # EDR without scaling
    edr = EffectiveDimensionalityReduction(GaussianProcessRegressor(),
                                           SVDTransformer(), normalize=False)
    X_scaled = StandardScaler().fit_transform(X)
    x2 = edr.fit_transform(X_scaled, y)

    assert np.allclose(x1, x2)


# @pytest.mark.parametrize("n_components,step", [(3, 1), (None, 0.99)])
# def test_iterative(n_components, step):
#     X = get_beta_inputs(300, 10)
#     B = np.linalg.qr(random_sparse(10, 3, density=0.3, random_state=0).A)[0]
#     y = get_edr_target(X.dot(B), 0.1)

#     gp_model = GaussianProcessRegressor(['RBF'], [{'ARD': True}])
#     edr = EffectiveDimensionalityReduction(gp_model,
#                                            SVDTransformer(),
#                                            n_components=n_components,
#                                            step=step, normalize=False)
#     edr.fit(X, y)
#     assert discrepancy(B, edr.components_.T) < 1e-1
#     assert edr.components_.shape[0] == 3
#     assert np.all(edr.subspace_variance_ratio_ == np.sort(edr.subspace_variance_ratio_)[::-1])


# @pytest.mark.parametrize("normalize,preprocessor",
#                          [(False, None),
#                           (True, None),
#                           (True, PCA(n_components=5))])
# def test_get_gradients_and_transform(normalize, preprocessor):
#     X = get_beta_inputs(300, 10)
#     B = np.linalg.qr(random_sparse(10, 3, density=0.3, random_state=0).A)[0]
#     y = get_edr_target(X.dot(B), 0.1)

#     gp_model = GaussianProcessRegressor(['RBF'], [{'ARD': True}])
#     edr = EffectiveDimensionalityReduction(gp_model,
#                                            SVDTransformer(),
#                                            step=2,
#                                            n_components=3,
#                                            normalize=normalize,
#                                            preprocessor=preprocessor)
#     edr.fit(X, y)
#     X_transform = edr.transform(X)
#     grads = edr.get_estimator_gradients(X)
#     assert grads.shape == X.shape
#     assert X_transform.shape == (300, 3)


# @pytest.mark.parametrize("normalize,preprocessor",
#                          [(False, None),
#                           (True, None),
#                           (True, PCA(n_components=5))])
# def test_refit(normalize, preprocessor):
#     X = get_beta_inputs(300, 10)
#     B = np.linalg.qr(random_sparse(10, 3, density=0.3, random_state=0).A)[0]
#     y = get_edr_target(X.dot(B), 0.1)

#     gp_model = GaussianProcessRegressor(['RBF'], [{'ARD': True}])
#     edr = EffectiveDimensionalityReduction(gp_model,
#                                            SVDTransformer(),
#                                            step=0.99,
#                                            normalize=normalize,
#                                            preprocessor=preprocessor)
#     edr.fit(X, y)
#     edr.refit(SparsePCA(n_components=3))
#     X_transform_refited = edr.transform(X, refitted=True)
#     assert X_transform_refited.shape == (300, 3)

# def test_block():
#     X = get_beta_inputs(300, 10)
#     B = np.linalg.qr(random_sparse(10, 3, density=0.4, random_state=0).A)[0]
#     B[:5, :2] = 0
#     B[5:, 2:] = 0
#     y = get_edr_target(X.dot(B), 0.1)

#     gp_model = GaussianProcessRegressor(['RBF'], [{'ARD': True}])
#     edr = BlockEDR(gp_model,
#                    SVDTransformer(),
#                    n_components=[2, 2], blocks=[range(5), range(5, 10)])
#     edr.fit(X, y)
#     X_transform = edr.transform(X)
#     edr.refit(SparsePCA(n_components=3))
#     X_transform_refited = edr.transform(X, refitted=True)
#     assert X_transform_refited.shape == (300, 3)
#     assert X_transform.shape == (300, 4)
#     assert np.all(edr.components_.T[5:, :2] == 0)
#     assert np.all(edr.components_.T[:5, 2:4] == 0)

# @pytest.mark.parametrize("normalize,preprocessor",
#                          [(False, None),
#                           (True, None),
#                           (True, PCA(n_components=5))])
# def test_block_preprocessing(normalize, preprocessor):
#     X = get_beta_inputs(300, 10)
#     B = np.linalg.qr(random_sparse(10, 3, density=0.4, random_state=0).A)[0]
#     B[:5, :2] = 0
#     B[5:, 2:] = 0
#     y = get_edr_target(X.dot(B), 0.1)

#     gp_model = GaussianProcessRegressor(['RBF'], [{'ARD': True}])
#     edr = EffectiveDimensionalityReduction(gp_model,
#                                            SVDTransformer(),
#                                            blocks=None,
#                                            n_components=[2, 2],
#                                            normalize=normalize,
#                                            preprocessor=preprocessor)
#     edr.fit(X, y)
#     X_transform = edr.transform(X)
#     grads = edr.get_estimator_gradients(X)
#     assert grads.shape == X.shape
#     assert X_transform.shape == (300, 4)

